@@ -3,9 +3,9 @@ use std::io::Write;
 use std::path::{Path, PathBuf};
 
 use anyhow::{ensure, Context, Result};
-use bellperson::bls::{Bls12, Fr};
 use bellperson::groth16;
 use bincode::{deserialize, serialize};
+use blstrs::{Bls12, Scalar as Fr};
 use filecoin_hashers::{Domain, Hasher};
 use log::{info, trace};
 use memmap::MmapOptions;
@@ -925,8 +925,10 @@ pub fn verify_seal<Tree: 'static + MerkleTreeTrait>(
     proof_vec: &[u8],
 ) -> Result<bool> {
     info!("verify_seal:start: {:?}", sector_id);
+
     ensure!(comm_d_in != [0; 32], "Invalid all zero commitment (comm_d)");
     ensure!(comm_r_in != [0; 32], "Invalid all zero commitment (comm_r)");
+    ensure!(!proof_vec.is_empty(), "Invalid proof bytes (empty vector)");
 
     let comm_r: <Tree::Hasher as Hasher>::Domain = as_safe_commitment(&comm_r_in, "comm_r")?;
     let comm_d: DefaultPieceDomain = as_safe_commitment(&comm_d_in, "comm_d")?;
@@ -1042,6 +1044,9 @@ pub fn verify_batch_seal<Tree: 'static + MerkleTreeTrait>(
             "Invalid all zero commitment (comm_r)"
         );
     }
+    for proofs in proof_vecs {
+        ensure!(!proofs.is_empty(), "Invalid proof (empty bytes) found");
+    }
 
     let sector_bytes = PaddedBytesAmount::from(porep_config);
 
